/*
 * Copyright (c) 2010, University of Michigan
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs and the author appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE VANDERBILT UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE VANDERBILT
 * UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE VANDERBILT UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE VANDERBILT UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: Thomas Schmid
 */

#include "cc2520.h"       // internal definitions
#include "cc2520driver.h" // public functions
#include "config.h"
#include "radio_alarm.h"
#include <stddef.h>
#include <string.h>
#include "../CoreSPI/spi.h"
#include <stdio.h>

//#define RADIO_DEBUG_MESSAGES

cc2520_header_t* getHeader(message_t* msg)
{
	return (cc2520_header_t*)(msg);
}

/*
 * Return a pointer to the data portion of the message.
 */
void* getPayload(message_t* msg)
{
	return ((void*)msg)  + sizeof(cc2520_header_t);
}

cc2520_metadata_t* getMeta(message_t* msg)
{
	return ((void*)msg) + sizeof(message_t) - sizeof(cc2520_metadata_t);
}

/*----------------- STATE -----------------*/

typedef enum
{
	STATE_VR_ON = 0,
	STATE_PD = 1,
	STATE_PD_2_IDLE = 2,
	STATE_IDLE = 3,
	STATE_IDLE_2_RX_ON = 4,
	STATE_RX_ON = 5,
	STATE_BUSY_TX_2_RX_ON = 6,
	STATE_IDLE_2_TX_ON = 7,
	STATE_TX_ON = 8,
	STATE_RX_DOWNLOAD = 9,
	STATE_SHUTDOWN = 10,
} radio_state_t;

radio_state_t state = STATE_VR_ON;

typedef enum
{
	CMD_NONE = 0,           // the state machine has stopped
	CMD_TURNOFF = 1,        // goto SLEEP state
	CMD_STANDBY = 2,        // goto TRX_OFF state
	CMD_SHUTDOWN = 3,      // turns the radio completely off
	CMD_TURNON = 4,         // goto RX_ON state
	CMD_TRANSMIT = 5,       // currently transmitting a message
	CMD_RECEIVE = 6,        // currently receiving a message
	CMD_CCA = 7,            // performing clear channel assessment
	CMD_CHANNEL = 8,        // changing the channel
	CMD_SIGNAL_DONE = 9,    // signal the end of the state transition
	CMD_DOWNLOAD = 10,       // download the received message
} radio_cmd_t;
radio_cmd_t cmd = CMD_NONE;

uint8_t radioIrq = 0;

uint8_t txPower;
uint8_t channel;

message_t* rxMsg;
message_t rxMsgBuffer;
message_t txMsgBuffer;
radio_message_t rxMessageBuffer;

#ifdef RADIO_DEBUG_MESSAGES
message_t* txMsg;
#endif

uint16_t capturedTime;  // the current time when the last interrupt has occured

uint8_t rssiClear;
uint8_t rssiBusy;


enum
{ // FIXME: need to check these for CC2520
	TX_SFD_DELAY = (uint16_t)(0 * RADIO_ALARM_MICROSEC),
	RX_SFD_DELAY = (uint16_t)(0 * RADIO_ALARM_MICROSEC),
};


inline cc2520_status_t getStatus();

void RadioAlarm_fired()
{

	if( state == STATE_PD_2_IDLE ) {
		state = STATE_IDLE;
		if( cmd == CMD_STANDBY )
			cmd = CMD_SIGNAL_DONE;
	}
	else if( state == STATE_IDLE_2_RX_ON ) {
		state = STATE_RX_ON;
		// in receive mode, enable SFD capture
		SfdCapture_captureRisingEdge();

		cmd = CMD_SIGNAL_DONE;
	}
	else
		ASSERT(FALSE);

	// make sure the rest of the command processing is called
	run();
}

/*----------------- REGISTER -----------------*/

inline cc2520_status_t writeRegister(uint8_t reg, uint8_t value)
{
	cc2520_status_t status;
	uint8_t v;

	ASSERT( SpiResource_isOwner() );

	CSN_set();
	CSN_clr();

	if( reg <= CC2520_FREG_MASK)
	{
		// we can use 1 byte less to write this register using the
		// register write command

		ASSERT( reg == (reg & CC2520_FREG_MASK) );


		status.value = SpiByte_write(CC2520_CMD_REGISTER_WRITE | reg);

	}
	else
	{
		// we have to use the memory write command as the register is in
		// SREG

		ASSERT( reg == (reg & CC2520_SREG_MASK) );

		// the register has to be below the 0x100 memory address. Thus, we
		// don't have to add anything to the MEMORY_WRITE command.
		status.value = SpiByte_write(CC2520_CMD_MEMORY_WRITE);
		status.value = SpiByte_write(reg);

	}
	// v is the value previously in the register
	v = SpiByte_write(value);

	CSN_set();

	return status;

}

/*
 * Strobes changed a lot between CC2420 and CC2520. They are now just an
 * other command, without any parameters.
 */
inline cc2520_status_t strobe(uint8_t reg)
{
	cc2520_status_t status;

	ASSERT( SpiResource_isOwner() );

	CSN_set();
	CSN_clr();

	status.value = SpiByte_write(reg);

	CSN_set();
	return status;

}

inline cc2520_status_t getStatus() {
	return strobe(CC2520_CMD_SNOP);
}

inline uint8_t readRegister(uint8_t reg)
{
	uint8_t value = 0;

	ASSERT( SpiResource_isOwner() );

	CSN_set();
	CSN_clr();

	if( reg <= CC2520_FREG_MASK )
	{
		ASSERT( reg == (reg & CC2520_FREG_MASK) );


		SpiByte_write(CC2520_CMD_REGISTER_READ | reg);

	}
	else
	{
		ASSERT( reg == (reg & CC2520_SREG_MASK) );

		SpiByte_write(CC2520_CMD_MEMORY_WRITE);
		SpiByte_write(reg);
	}

	value = SpiByte_write(0);
	CSN_set();

	return value;
}

inline cc2520_status_t writeTxFifo(uint8_t* data, uint8_t length)
{
	cc2520_status_t status;
	uint8_t idx;

	ASSERT( SpiResource_isOwner() );

	CSN_set();
	CSN_clr();

	status.value = SpiByte_write(CC2520_CMD_TXBUF);
	// FIXME: replace this at some point with a SPIPacket call.
	for(idx = 0; idx<length; idx++)
		SpiByte_write(data[idx]);

	CSN_set();

	return status;

}


inline uint8_t waitForRxFifoNoTimeout() {
	// wait for fifo to go high
	while(FIFO_get() == 0);

	return FIFO_get();
}

inline uint8_t waitForRxFifo() {
	// wait for fifo to go high or timeout
	// timeout is now + 2 byte time (4 symbol time)
	uint16_t timeout = RadioAlarm_getNow() + 4 * CC2520_SYMBOL_TIME;

	while(FIFO_get() == 0 && (timeout - RadioAlarm_getNow() < 0x7FFF));
	return FIFO_get();
}

inline cc2520_status_t readLengthFromRxFifo(uint8_t* lengthPtr)
{
	cc2520_status_t status;

	ASSERT( SpiResource_isOwner() );
	ASSERT( CSN_get() == 1 );

	// FIXME: ???? why do we do this ????
	CSN_set();
	CSN_clr();
	CSN_set();
	CSN_clr();
	CSN_set();
	CSN_clr();

	status.value = SpiByte_write(CC2520_CMD_RXBUF);
	waitForRxFifoNoTimeout();
	*lengthPtr = SpiByte_write(0);
	return status;
}

inline void readPayloadFromRxFifo(uint8_t* data, uint8_t length)
{
	uint8_t idx;

	// readLengthFromRxFifo was called before, so CSN is cleared and spi is ours
	ASSERT( CSN_get() == 0 );

	for(idx = 0; idx<length; idx++) {
		waitForRxFifo();
		ASSERT(FIFO_get());
		data[idx] = SpiByte_write(0);
	}
}

inline void readRssiFromRxFifo(uint8_t* rssiPtr)
{
	// FIXME: make sure that RSSI is added to the frame in the
	// configuration! See 20.3.4 in CC2520 Manual (Dec. 2007)

	// readLengthFromRxFifo was called before, so CSN is cleared and spi is ours

	waitForRxFifo();
	ASSERT(FIFO_get());
	*rssiPtr = SpiByte_write(0);
}

inline void readCrcOkAndLqiFromRxFifo(uint8_t* crcOkAndLqiPtr)
{
	// readLengthFromRxFifo was called before, so CSN is cleared and spi is ours
	ASSERT( CSN_get() == 0 );

	*crcOkAndLqiPtr = SpiByte_write(0);

	// end RxFifo read operation
	CSN_set();
}

inline void flushRxFifo() {
	// set it to stop possible pending fifo transfer

	{
		cc2520_status_t status = strobe(CC2520_CMD_SFLUSHRX);
#ifdef RADIO_DEBUG_MESSAGES
		printf("b_flush %u\n\r", status.value);
#else
		status.value = 0;
#endif
	}
	// FIXME: Why so many? Bug in CC2420?
	/*
//      strobe(CC2420X_SRFOFF);
	strobe(CC2420X_SFLUSHRX);
	strobe(CC2420X_SFLUSHRX);
	strobe(CC2420X_SFLUSHRX);
	{
		cc2420X_status_t status = strobe(CC2420X_SFLUSHRX);
#ifdef RADIO_DEBUG_MESSAGES
		if( call DiagMsg.record() )
		{
			call DiagMsg.str("a_flush");
			call DiagMsg.uint16(status.value);
			call DiagMsg.send();
		}
#endif
*/
	//strobe(CC2520_CMD_SRXON);


}

/*----------------- INIT -----------------*/

void pre_init_radio()
{
	// CSN is active low
	CSN_set();

	// start up voltage regulator
	VREN_clr();
	VREN_set();

	// do a reset
	RSTN_clr();
	// hold line low for Tdres
	busywait( 200 ); // typical .1ms VR startup time

	RSTN_set();
	// wait another .2ms for xosc to stabilize
	busywait( 200 );


	rxMsg = &rxMsgBuffer;

	state = STATE_VR_ON;
}

void Radio_init()
{
	// initialize the radio timer
	RadioAlarm_init(&RadioAlarm_fired);

	MSS_GPIO_init();

	// set pin directions
	MSS_GPIO_config( FIFO, MSS_GPIO_INPUT_MODE); // FIFO
	MSS_GPIO_config( FIFOP, MSS_GPIO_INPUT_MODE | MSS_GPIO_IRQ_EDGE_POSITIVE ); // FIFOP
	MSS_GPIO_config( CCA, MSS_GPIO_INPUT_MODE); // CCA
	MSS_GPIO_config( NONE, MSS_GPIO_INPUT_MODE); // none
	MSS_GPIO_config( VREN, MSS_GPIO_OUTPUT_MODE);// vreg_en
	MSS_GPIO_config( RSTN, MSS_GPIO_OUTPUT_MODE);// reset_n
	MSS_GPIO_config( CLK, MSS_GPIO_INPUT_MODE); // clk
	MSS_GPIO_config( SFD, MSS_GPIO_INPUT_MODE | MSS_GPIO_IRQ_EDGE_POSITIVE ); // sfd

	MSS_GPIO_disable_irq( FIFOP );
	MSS_GPIO_disable_irq( SFD );

	// Init SPI
	spi_init();

	pre_init_radio();

	initRadio();
}

inline void resetRadio() {
	// now register access is enabled: set up defaults
	cc2520_fifopctrl_t fifopctrl;
	cc2520_frmfilt0_t frmfilt0;
	cc2520_frmfilt1_t frmfilt1;
	cc2520_srcmatch_t srcmatch;

	// do a reset
	RSTN_clr();
	busywait( 200 ); //
	RSTN_set();

	// update default values of registers
	// given from SWRS068, December 2007, Section 28.1
	writeRegister(CC2520_TXPOWER, cc2520_txpower_default.value);
	writeRegister(CC2520_CCACTRL0, cc2520_ccactrl0_default.value);
	writeRegister(CC2520_MDMCTRL0, cc2520_mdmctrl0_default.value);
	writeRegister(CC2520_MDMCTRL1, cc2520_mdmctrl1_default.value);
	writeRegister(CC2520_RXCTRL, cc2520_rxctrl_default.value);
	writeRegister(CC2520_FSCTRL, cc2520_fsctrl_default.value);
	writeRegister(CC2520_FSCAL1, cc2520_fscal1_default.value);
	writeRegister(CC2520_AGCCTRL1, cc2520_agcctrl1_default.value);
	writeRegister(CC2520_ADCTEST0, cc2520_adctest0_default.value);
	writeRegister(CC2520_ADCTEST1, cc2520_adctest1_default.value);
	writeRegister(CC2520_ADCTEST2, cc2520_adctest2_default.value);

	// setup fifop threshold
	fifopctrl.f.fifop_thr = 127;
	writeRegister(CC2520_FIFOPCTRL, fifopctrl.value);

	// FIXME: disable frame filtering for now
	frmfilt0 = cc2520_frmfilt0_default;
	frmfilt0.f.frame_filter_en = 0;
	writeRegister(CC2520_FRMFILT0, frmfilt0.value);

	// accept reserved frames
	frmfilt1 = cc2520_frmfilt1_default;
	frmfilt1.f.accept_ft_4to7_reserved = 1;
	writeRegister(CC2520_FRMFILT1, frmfilt1.value);

	// disable src address decoding
	srcmatch = cc2520_srcmatch_default;
	srcmatch.f.src_match_en = 0;
	writeRegister(CC2520_SRCMATCH, srcmatch.value);

	// enable auto crc and append rssi to frame
	// this is done by default.
}

void initRadio()
{
	resetRadio();

	txPower = CC2520_DEF_RFPOWER & CC2520_TX_PWR_MASK;
	channel = CC2520_DEF_CHANNEL & CC2520_CHANNEL_MASK;

	state = STATE_PD;
}

/*----------------- CHANNEL -----------------*/

uint8_t getChannel()
{
	return channel;
}

error_t Radio_setRadioChannel(uint8_t c)
{
	c &= CC2520_CHANNEL_MASK;
	if(c < 11 || c > 26)
		return EINVAL;
	if( cmd != CMD_NONE )
		return EBUSY;
	else if( channel == c )
		return EALREADY;

	channel = c;
	cmd = CMD_CHANNEL;
	run();

	return SUCCESS;
}

inline void setChannel()
{
	cc2520_freqctrl_t freqctrl;
	// set up freq
	freqctrl = cc2520_freqctrl_default;
	freqctrl.f.freq = 11 + 5 * (channel - 11);
#ifdef RADIO_DEBUG_MESSAGES
	printf("freqctrl %u\n\r", freqctrl.value);
#endif

	writeRegister(CC2520_FREQCTRL, freqctrl.value);
}

inline void changeChannel()
{
	ASSERT( cmd == CMD_CHANNEL );
	ASSERT( state == STATE_PD || state == STATE_IDLE || ( state == STATE_RX_ON && RadioAlarm_isFree()));

	if( isSpiAcquired() )
	{
		setChannel();

		if( state == STATE_RX_ON ) {
			RadioAlarm_wait(IDLE_2_RX_ON_TIME); // 12 symbol periods
			state = STATE_IDLE_2_RX_ON;
		}
		else
			cmd = CMD_SIGNAL_DONE;
	}
}

/*----------------- TURN ON/OFF -----------------*/

inline void changeState()
{
	if( (cmd == CMD_STANDBY || cmd == CMD_TURNON)
			&& state == STATE_SHUTDOWN && isSpiAcquired() && RadioAlarm_isFree() )
	{
		pre_init_radio(); // startup VR
		resetRadio();

		state = STATE_PD;
	}
	if( (cmd == CMD_STANDBY || cmd == CMD_TURNON)
		&& state == STATE_PD  && isSpiAcquired() && RadioAlarm_isFree() )
	{

		// start oscillator
		strobe(CC2520_CMD_SXOSCON);

		RadioAlarm_wait(PD_2_IDLE_TIME); // .86ms OSC startup time
		state = STATE_PD_2_IDLE;
	}
	else if( cmd == CMD_TURNON && state == STATE_IDLE && isSpiAcquired() && RadioAlarm_isFree())
	{
		// setChannel was ignored in SLEEP because the SPI was not working, so do it here
		setChannel();

		// start receiving
		strobe(CC2520_CMD_SRXON);
		RadioAlarm_wait(IDLE_2_RX_ON_TIME); // 12 symbol periods
		state = STATE_IDLE_2_RX_ON;
		strobe(CC2520_CMD_SNOP);
	}
	else if( (cmd == CMD_TURNOFF || cmd == CMD_STANDBY || cmd == CMD_SHUTDOWN)
		&& state == STATE_RX_ON && isSpiAcquired() )
	{
		// stop receiving
		strobe(CC2520_CMD_SRFOFF);

		state = STATE_IDLE;
	}

	if( cmd == CMD_TURNOFF && state == STATE_IDLE  && isSpiAcquired() )
	{
		// stop oscillator
		strobe(CC2520_CMD_SXOSCOFF);

		// do a reset
		resetRadio();
		state = STATE_PD;
		cmd = CMD_SIGNAL_DONE;
	}
	else if( cmd == CMD_STANDBY && state == STATE_IDLE )
		cmd = CMD_SIGNAL_DONE;
	else if( cmd == CMD_SHUTDOWN && state == STATE_IDLE)
	{
		// shutdown radio
		CSN_set();
		VREN_clr();
		state = STATE_SHUTDOWN;
		cmd = CMD_NONE;
	}
}

// TODO: turn off SFD capture when turning off radio
error_t RadioState_turnOff()
{
	if( cmd != CMD_NONE )
		return EBUSY;
	else if( state == STATE_PD )
		return EALREADY;

#ifdef RADIO_DEBUG_MESSAGES
	printf("turnOff\n\r");
#endif

	cmd = CMD_TURNOFF;
	run();

	return SUCCESS;
}

error_t RadioState_standby()
{
	if( cmd != CMD_NONE || (state == STATE_PD && ! RadioAlarm_isFree()) )
		return EBUSY;
	else if( state == STATE_IDLE )
		return EALREADY;

#ifdef RADIO_DEBUG_MESSAGES
	printf("standBy\n\r");
#endif

	cmd = CMD_STANDBY;
	run();
	return SUCCESS;
}

error_t RadioState_shutdown()
{
	if( cmd != CMD_NONE || (state == STATE_PD && ! RadioAlarm_isFree()) )
		return EBUSY;
	else if( state == STATE_SHUTDOWN )
		return EALREADY;

#ifdef RADIO_DEBUG_MESSAGES
	printf("Shutdown\n\r");
#endif

	cmd = CMD_SHUTDOWN;
	run();
	return SUCCESS;

}
// TODO: turn on SFD capture when turning off radio
error_t RadioState_turnOn()
{
	if( cmd != CMD_NONE || (state == STATE_PD && ! RadioAlarm_isFree()) )
		return EBUSY;
	else if( state == STATE_RX_ON )
		return EALREADY;

#ifdef RADIO_DEBUG_MESSAGES
	printf("turnOn\n\r");
#endif

	cmd = CMD_TURNON;
	run();

	return SUCCESS;
}

/*----------------- TRANSMIT -----------------*/

error_t RadioSend_send(message_t* msg)
{
	uint16_t time;
	uint8_t p;
	uint8_t length;
	uint8_t* data;
	uint8_t header;
	//uint32_t time32;
	//void* timesync;
	cc2520_status_t status;


	if( cmd != CMD_NONE || (state != STATE_IDLE && state != STATE_RX_ON) || ! isSpiAcquired() || radioIrq )
		return EBUSY;

	p = (CC2520_DEF_RFPOWER) & CC2520_TX_PWR_MASK;

	if( p != txPower )
	{
		cc2520_txpower_t txpower = cc2520_txpower_default;

		txPower = p;

		txpower.f.pa_power = txPower;
		writeRegister(CC2520_TXPOWER, txpower.value);
	}

#ifdef RADIO_DEBUG_MESSAGES
	{
		uint8_t tmp2;
		tmp2 = CCA_get();
		length = getHeader(msg)->length;
		printf("cca %u\n\r", tmp2);
	}
#else
	//FIXME: we currently ignore CCA
	//if( call Config.requiresRssiCca(msg) && !call CCA.get() )
	//	return EBUSY;
#endif

	// there's a chance that there was a receive SFD interrupt in such a
	// short time.
	// TODO: there's still a chance

	if (SFD_get() == 1 || radioIrq)
		return EBUSY;
	else
		// stop receiving
		strobe(CC2520_CMD_SRFOFF);

	ASSERT( ! radioIrq );

	data = getPayload(msg);
	length = getHeader(msg)->length;

	// length | data[0] ... data[length-3] | automatically generated FCS

	__disable_irq();
	writeTxFifo(&length, 1);
	__enable_irq();

	// FCS is automatically generated
	length -= 2;

	// preload fcf, dsn, destpan, and dest
	header = 7;
	if( header > length )
		header = length;

	length -= header;

	// first upload the header to gain some time
	__disable_irq();
	writeTxFifo(data, header);
	__enable_irq();

	__disable_irq();
	{
		strobe(CC2520_CMD_STXON);
		time = RadioAlarm_getNow();
		SfdCapture_captureFallingEdge();
		state = STATE_TX_ON;
	}
	__enable_irq();

#ifdef RADIO_DEBUG_MESSAGES
	txMsg = msg;
#endif

	// do something useful, just to wait a little (12 symbol periods)
	//time32 = call LocalTime.get();
	//timesync = call PacketTimeSyncOffset.isSet(msg) ? ((void*)msg) + call PacketTimeSyncOffset.get(msg) : 0;

	//time32 += (int16_t)(time + TX_SFD_DELAY) - (int16_t)(time32);

	//if( timesync != 0 )
	//	*(timesync_relative_t*)timesync = (*(timesync_absolute_t*)timesync) - time32;

	// write the rest of the payload to the fifo
	__disable_irq();
	writeTxFifo(data+header, length);
	__enable_irq();

	// get status
	status = getStatus();
	ASSERT ( status.tx_active == 1);
	// FIXME: have to check for underflow exception!
	//ASSERT ( status.tx_underflow == 0);
	ASSERT ( status.xosc_stable == 1);

	//if( timesync != 0 )
	//	*(timesync_absolute_t*)timesync = (*(timesync_relative_t*)timesync) + time32;

	//call PacketTimeStamp.set(msg, time32);

	// wait for SFD falling edge
	state = STATE_BUSY_TX_2_RX_ON;
	cmd = CMD_TRANSMIT;

#ifdef RADIO_DEBUG_MESSAGES
	{
		uint8_t i;
		char* payload = (char*)getPayload(msg);

		length = getHeader(msg)->length;

		printf("t %u\n\r  ", length);
		for(i=0; i< length-2; i++)
			printf("0x%x ", payload[i]);
		printf("\n\r");
	}
#endif


	return SUCCESS;
}

/*----------------- CCA -----------------*/

error_t Radio_requestCCA()
{
	if( cmd != CMD_NONE || state != STATE_RX_ON )
		return EBUSY;

	if(CCA_get()) {
		return SUCCESS;
	} else {
		return EBUSY;
	}
}

/*----------------- RECEIVE -----------------*/

// recover from an error
// rx fifo flush does not always work
inline void recover() {
	cc2520_status_t status;

	// reset the radio, initialize registers to default values
	ASSERT(0);
	resetRadio();

	SfdCapture_disable();

	ASSERT(state == STATE_PD);

	// start oscillator
	strobe(CC2520_CMD_SXOSCON);

	// going idle in PD_2_IDLE_TIME
	state = STATE_PD_2_IDLE;

	busywait(PD_2_IDLE_TIME); // .86ms OSC startup time

	// get status
	status = getStatus();
	ASSERT ( status.rssi_valid == 0);
	//ASSERT ( status.lock == 0);
	ASSERT ( status.tx_active == 0);
	//ASSERT ( status.enc_busy == 0);
	//ASSERT ( status.tx_underflow == 0);
	ASSERT ( status.xosc_stable == 1);

	// we're idle now
	state = STATE_IDLE;

	// download current channel to the radio
	setChannel();

	// start receiving
	strobe(CC2520_CMD_SRXON);
	state = STATE_IDLE_2_RX_ON;

	SfdCapture_captureRisingEdge();

	// we will be able to receive packets in 12 symbol periods
	state = STATE_RX_ON;
}

inline void downloadMessage()
{
	uint8_t length;
	uint16_t crc = 1;
	uint8_t* data;
	uint8_t rssi;
	uint8_t crc_ok_lqi;
	//uint16_t sfdTime;

	state = STATE_RX_DOWNLOAD;

	//atomic sfdTime = capturedTime;

	// data starts after the length field
	data = getPayload(rxMsg);

	// read the length byte
	readLengthFromRxFifo(&length);

#ifdef RADIO_DEBUG_MESSAGES
	{
		uint8_t i;
		char* payload = (char*)getPayload(rxMsg);
		printf("rx %d\n\r   ", length);
		// the last two bytes are FCS/CRC
		for(i=0; i<length-2; i++)
			printf("0x%x ", payload[i]);
		printf("\n\r");
	}
#endif

	// check for too short lengths
	if (length == 0) {
		// stop reading RXFIFO
		CSN_set();

		ASSERT( FIFOP_get() == 0 );
		ASSERT( FIFO_get() == 0 );

		state = STATE_RX_ON;
		cmd = CMD_NONE;
		SfdCapture_captureRisingEdge();
		return;
	}

	if (length == 1) {
		// skip payload and rssi
		__disable_irq();
		readCrcOkAndLqiFromRxFifo(&crc_ok_lqi);
		__enable_irq();

		ASSERT( FIFOP_get() == 0 );
		ASSERT( FIFO_get() == 0 );

		state = STATE_RX_ON;
		cmd = CMD_NONE;
		SfdCapture_captureRisingEdge();
		return;
	}

	if (length == 2) {
		// skip payload
		__disable_irq();
		readRssiFromRxFifo(&rssi);
		readCrcOkAndLqiFromRxFifo(&crc_ok_lqi);
		__enable_irq();

		ASSERT( FIFOP_get() == 0 );
		ASSERT( FIFO_get() == 0 );

		state = STATE_RX_ON;
		cmd = CMD_NONE;
		SfdCapture_captureRisingEdge();
		return;
	}

	// check for too long lengths

	if( length > 127 ) {

		__disable_irq();
		recover();
		__enable_irq();

		ASSERT( FIFOP_get() == 0 );
		ASSERT( FIFO_get() == 0 );

		state = STATE_RX_ON;
		cmd = CMD_NONE;
		SfdCapture_captureRisingEdge();
		return;
	}

	if( length > MAX_PAYLOAD_LENGTH )
	{
		while( length-- > 2 ) {
			__disable_irq();
			readPayloadFromRxFifo(data, 1);
			__enable_irq();
		}

		__disable_irq();
		readRssiFromRxFifo(&rssi);
		readCrcOkAndLqiFromRxFifo(&crc_ok_lqi);
		__enable_irq();

		ASSERT( FIFOP_get() == 0 );

		state = STATE_RX_ON;
		cmd = CMD_NONE;
		SfdCapture_captureRisingEdge();
		return;
	}

	// if we're here, length must be correct
	ASSERT(length >= 3 && length <= MAX_PAYLOAD_LENGTH);

	getHeader(rxMsg)->length = length;

	// we'll read the FCS/CRC separately
	length -= 2;

	// download the whole payload
	readPayloadFromRxFifo(data, length );

	// the last two bytes are not the fsc, but RSSI(8), CRC_ON(1)+LQI(7)
	readRssiFromRxFifo(&rssi);
	readCrcOkAndLqiFromRxFifo(&crc_ok_lqi);

	// there are still bytes in the fifo or if there's an overflow,
	// recover
	// TODO: actually, we can signal that a message was received, without
	// timestamp set
	if (FIFOP_get() == 1 || FIFO_get() == 1) {
		__disable_irq();
		recover();
		__enable_irq();
		state = STATE_RX_ON;
		cmd = CMD_NONE;
		SfdCapture_captureRisingEdge();
		return;
	}

	state = STATE_RX_ON;
	cmd = CMD_NONE;

	// ready to receive new message: enable SFD interrupts
	SfdCapture_captureRisingEdge();

	if( RadioReceive_header(rxMsg) )
	{
		// set RSSI, CRC and LQI only if we're accepting the message
		//((cc2520_metadata_t*)(rxMsg->metadata))->rssi = rssi;
		getMeta(rxMsg)->rssi = rssi;
		//((cc2520_metadata_t*)(rxMsg->metadata))->lqi = crc_ok_lqi & 0x7f;
		getMeta(rxMsg)->lqi = crc_ok_lqi & 0x7f;
		crc = (crc_ok_lqi > 0x7f) ? 0 : 1;
	}

	// signal only if it has passed the CRC check
	if( crc == 0 )
	{
	   //uint32_t time32;

		//time32 = call LocalTime.get();
		//atomic time32 += (int16_t)(sfdTime - RX_SFD_DELAY) - (int16_t)(time32);
		//call PacketTimeStamp.set(rxMsg, time32);

#ifdef RADIO_DEBUG_MESSAGES
		{
			uint8_t i;
			char* payload = (char*)getPayload(rxMsg);
			printf("rx %d\n\r   ", length);
			for(i=0; i<length; i++)
				printf("0x%x ", payload[i]);
			printf("\n\r");
		}
#endif
		rxMsg = RadioReceive_receive(rxMsg);

	}

}

/*----------------- IRQ -----------------*/

// RX SFD (rising edge), disabled for TX
__attribute__ ((interrupt)) void GPIO7_IRQHandler( void )
{
    MSS_GPIO_clear_irq( SFD );
    NVIC_ClearPendingIRQ( GPIO7_IRQn );

	ASSERT( ! radioIrq );
	ASSERT( state == STATE_RX_ON || state == STATE_TX_ON || state == STATE_BUSY_TX_2_RX_ON);

	radioIrq = TRUE;
	SfdCapture_disable();
	//capturedTime = time;

#ifdef RADIO_DEBUG_MESSAGES
	{
		printf("SFD s=%u", state);
		if(FIFO_get())
			printf(" FIFO");
		if(FIFOP_get())
			printf(" FIFOP");
		if(SFD_get())
			printf(" SFD");
		printf("\n\r");
	}
#endif

	// do the rest of the processing
	run();
}

// FIFOP interrupt, last byte received
__attribute__ ((interrupt)) void GPIO1_IRQHandler( void )
{
    MSS_GPIO_clear_irq( FIFOP );
    NVIC_ClearPendingIRQ( GPIO1_IRQn );
	// not used
}

inline void serviceRadio()
{
	__disable_irq();
	if( isSpiAcquired() )
	{
		radioIrq = FALSE;

		if( state == STATE_RX_ON && cmd == CMD_NONE )
		{
			// it's an RX SFD
			cmd = CMD_DOWNLOAD;
		}
		else if( state == STATE_BUSY_TX_2_RX_ON && cmd == CMD_TRANSMIT)
		{
			// it's a TX_END
			state = STATE_RX_ON;
			cmd = CMD_NONE;
#ifdef RADIO_DEBUG_MESSAGES
		{
			printf("txdone ");
			if(FIFO_get())
				printf(" FIFO");
			if(FIFOP_get())
				printf(" FIFOP");
			if(SFD_get())
				printf(" SFD");
			printf("\n\r");
		}
#endif

			SfdCapture_captureRisingEdge();
			strobe(CC2520_CMD_SRXON);
			Radio_sendDone(SUCCESS);

		}

		else
			ASSERT(FALSE);
	}
	__enable_irq();
}

uint8_t RadioReceive_header(message_t* msg)
{
	return TRUE;
}

error_t Radio_send(uint8_t seqn, uint16_t panid, uint16_t saddr, uint16_t daddr, uint8_t* payload, uint8_t length)
{
	radio_message_t m;
	cc2520_header_t* h = getHeader(&txMsgBuffer);
	uint8_t* p = getPayload(&txMsgBuffer);
	h->length = length + sizeof(radio_message_t) + 2; // +2 for FCS which is automatically generated
	m.fcf = 0x8841;
	m.seqn = seqn;
	m.panid = panid;
	m.saddr = saddr;
	m.daddr = daddr;
	memcpy(p, &m, sizeof(radio_message_t));
	if(length < MAX_PAYLOAD_LENGTH - sizeof(radio_message_t))
	{
		memcpy(p + sizeof(radio_message_t), payload, length);
		return RadioSend_send(&txMsgBuffer);
	} else {
		return FAIL;
	}
}

message_t* RadioReceive_receive(message_t* msg)
{
	radio_message_t *m;
	uint8_t* payload;
	cc2520_header_t* h = getHeader(msg);
	cc2520_metadata_t* meta = getMeta(msg);
	uint8_t length = h->length;

	m = (radio_message_t*)getPayload(msg);
	payload = getPayload(msg) + sizeof(radio_message_t);

	Radio_receive(m->seqn, m->panid, m->saddr, m->daddr, payload, length - sizeof(radio_message_t), (int8_t)meta->rssi);
	return msg;
}

/*----------------- TASKLET -----------------*/

void run()
{
	if( radioIrq )
		serviceRadio();

	if( cmd != CMD_NONE )
	{
		if( cmd == CMD_DOWNLOAD && state == STATE_RX_ON)
			downloadMessage();
		else if( CMD_TURNOFF <= cmd && cmd <= CMD_TURNON )
			changeState();
		else if( cmd == CMD_CHANNEL )
			changeChannel();

		if( cmd == CMD_SIGNAL_DONE )
		{
			cmd = CMD_NONE;
		}
	}

	//if( cmd == CMD_NONE && state == STATE_RX_ON && ! radioIrq )
	//	RadioSend_ready();

	if( cmd == CMD_NONE )
		SpiResource_release();
}

/*----------------- RadioConfig -----------------*/
uint8_t Config_headerLength(message_t* msg)
{
    return offsetof(message_t, data) - sizeof(cc2520_header_t);
}

/*----------------- RadioPacket -----------------*/

uint8_t RadioPacket_headerLength(message_t* msg)
{
	return Config_headerLength(msg) + sizeof(cc2520_header_t);
}

uint8_t RadioPacket_payloadLength(message_t* msg)
{
	return getHeader(msg)->length - 2;
}

void RadioPacket_setPayloadLength(message_t* msg, uint8_t length)
{
	ASSERT( 1 <= length && length <= 125 );
	ASSERT( RadioPacket_headerLength(msg) + length <= sizeof(message_t) );

	// we add the length of the CRC, which is automatically generated
	getHeader(msg)->length = length + 2;
}
